"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var esbasejs_1 = require("esbasejs");
var Ieee754FloatingPointUtilities = (function () {
    function Ieee754FloatingPointUtilities(mantNoBits, expNoBits) {
        if (!esbasejs_1.isIntegralDouble(mantNoBits) || mantNoBits < 0 || 52 < mantNoBits)
            throw new Error();
        if (!esbasejs_1.isIntegralDouble(mantNoBits) || expNoBits < 0 || 11 < expNoBits)
            throw new Error();
        this._mantNoBits = mantNoBits;
        this._expNoBits = expNoBits;
        this._expBias = (1 << (this._expNoBits - 1)) - 1;
        this._expMax = (1 << this._expNoBits) - 1 - this._expBias;
    }
    Ieee754FloatingPointUtilities.prototype.getMantissaSize_base2 = function () {
        return this._mantNoBits;
    };
    Ieee754FloatingPointUtilities.prototype.getExponentSize_base2 = function () {
        return this._expNoBits;
    };
    Ieee754FloatingPointUtilities.prototype.getExponentMin = function () {
        return -this._expBias;
    };
    Ieee754FloatingPointUtilities.prototype.getExponentMax = function () {
        return this._expMax;
    };
    Ieee754FloatingPointUtilities.prototype.getExponentBias = function () {
        return this._expBias;
    };
    return Ieee754FloatingPointUtilities;
}());
exports.Ieee754FloatingPointUtilities = Ieee754FloatingPointUtilities;
var pow = Math.pow;
var DoubleUtilities = (function (_super) {
    __extends(DoubleUtilities, _super);
    function DoubleUtilities() {
        _super.call(this, 52, 11);
        this._2PowMantNoBits = pow(2, this.getMantissaSize_base2());
        this._minNormalizedV = pow(2, this.getExponentMin() + 1);
        var arrE = [512, 256, 128, 64, 32, 16, 8, 4, 2, 1];
        this._arrE = arrE;
        var arrE_len = this._arrE.length;
        var arr2PowE = new Array(arrE_len);
        this._arr2PowE = arr2PowE;
        for (var i = 0; i < arrE_len; ++i) {
            arr2PowE[i] = pow(2, arrE[i]);
        }
        this._mantMax = this._2PowMantNoBits - 1;
    }
    DoubleUtilities.prototype.create = function (exponent, mantissa, isNegative) {
        if (isNegative === void 0) { isNegative = false; }
        var expMin;
        var expMax;
        if (exponent < (expMin = this.getExponentMin())
            || (expMax = this.getExponentMax()) < exponent
            || !esbasejs_1.isIntegralDouble(exponent)) {
            throw new Error();
        }
        if (mantissa < 0 || this._mantMax < mantissa || !esbasejs_1.isIntegralDouble(mantissa)) {
            throw new Error();
        }
        if (typeof isNegative !== "boolean") {
            throw new TypeError();
        }
        var n;
        if (exponent === expMax) {
            if (mantissa === 0) {
                n = 1 / 0;
            }
            else {
                n = 0 / 0;
            }
        }
        else {
            n = (mantissa + (exponent !== expMin ? this._2PowMantNoBits : 0)) * pow(2, exponent - this.getMantissaSize_base2());
        }
        if (isNegative) {
            n = -n;
        }
        return n;
    };
    DoubleUtilities.prototype.getExponent = function (v1) {
        if (typeof v1 === "number") {
            if (isFinite(v1)) {
                if (v1 < 0) {
                    v1 = -v1;
                }
                if (this._minNormalizedV <= v1) {
                    var v2 = void 0, i = 0, e = 0;
                    var arrE = this._arrE;
                    var n = arrE.length;
                    var arr2PowE = this._arr2PowE;
                    if (v1 < 1) {
                        for (; i < n; ++i) {
                            v2 = v1 * arr2PowE[i];
                            if (v2 < 2) {
                                v1 = v2;
                                e -= arrE[i];
                            }
                        }
                    }
                    else {
                        for (; i < n; ++i) {
                            v2 = arr2PowE[i];
                            if (v2 <= v1) {
                                v1 /= v2;
                                e += arrE[i];
                            }
                        }
                    }
                    return e;
                }
                return this.getExponentMin();
            }
            return this.getExponentMax();
        }
        throw new Error();
    };
    DoubleUtilities.getInstance = function () {
        return DoubleUtilities.instance;
    };
    DoubleUtilities.instance = new DoubleUtilities();
    return DoubleUtilities;
}(Ieee754FloatingPointUtilities));
exports.DoubleUtilities = DoubleUtilities;
// This operation is equivalent to floor(log2(v1)), but this implementation is guaranteed to be correct for all doubles (100% accuracy).
// Furthermore, this implementation throws an error for non-positive and non-finite v1 (for which log2 is mathematically undefined).
function double_log2Floor(x) {
    var exp, mant;
    if (!esbasejs_1.isFiniteDouble(x) || x <= 0)
        throw new Error();
    if (x < 2)
        return 0;
    exp = DoubleUtilities.getInstance().getExponent(x);
    return exp;
}
exports.double_log2Floor = double_log2Floor;
//# sourceMappingURL=ieee754.js.map